#  cli_driver.py
#  CLI driver class for the Hammer VLSI abstraction.
#
#  See LICENSE for licence details.

import argparse
import json
import os
import subprocess
import sys
import tempfile
from pathlib import Path
import importlib.resources

import ruamel.yaml

import hammer.config
from .hammer_vlsi_impl import HammerTool, HammerVLSISettings
from .hooks import HammerToolHookAction, HammerStartStopStep
from .driver import HammerDriver, HammerDriverOptions
from .hammer_build_systems import BuildSystems

from functools import reduce
from textwrap import dedent
from typing import List, Dict, Tuple, Any, Callable, Optional, Union, cast

from hammer.utils import add_dicts, deeplist, deepdict, get_or_else, check_function_type

from hammer.config import HammerJSONEncoder


KEY_DIR = tempfile.mkdtemp()
KEY_PATH = os.path.join(KEY_DIR, "key-history.json")

def parse_optional_file_list_from_args(args_list: Any, append_error_func: Callable[[str], None]) -> List[str]:
    """Parse a possibly null list of files, validate the existence of each file, and return a list of paths (possibly
    empty)."""
    results = []  # type: List[str]
    if args_list is None:
        # No arguments
        pass
    elif isinstance(args_list, List):
        # Check for file's existence before canonicalization so that we can give the user a sane error message
        for c in args_list:
            if not os.path.exists(c):
                append_error_func("Given path %s does not exist!" % c)
        # Canonicalize the path at this point
        results = list(map(os.path.realpath, args_list))
    else:
        append_error_func("Argument was not a list?")
    return results


def get_nonempty_str(arg: Any) -> Optional[str]:
    """Either get the non-empty string from the given arg or None if it is not a non-empty string."""
    if isinstance(arg, str):
        if len(arg) > 0:
            return str(arg)
    return None


def dump_config_to_json_file(output_path: str, config: dict) -> None:
    """
    Helper function to dump the given config to the given output path while overwriting it if it already exists.
    :param output_path: Output path (e.g. "obj/output.log")
    :param config: Config dictionary to dump
    """
    with open(output_path, "w") as f:
        f.write(json.dumps(config, cls=HammerJSONEncoder, indent=4))

def dump_config_to_yaml_file(output_path: str, config: dict) -> None:
    """
    Helper function to dump the given config in YAML form
    to the given output path while overwriting it if it already exists.
    :param output_path: Output path
    :param config: Config dictionary to dump
    """
    yaml = ruamel.yaml.YAML()
    yaml.indent(offset=2, sequence=4)
    with open(output_path, 'w') as f:
        yaml.dump(config, f)

def add_key_history(config: dict, history: dict) -> Any:
    """
    Generates a YAML file with comments indicating what files modified said keys.
    """
    new = ruamel.yaml.CommentedMap()
    curr_slot = 0
    for k, v in config.items():
        if k in history:
            pretty_hist = ', '.join(history[k])
            new.insert(curr_slot, k, v, comment=f"Modified by: {pretty_hist}")
        else:
            new.insert(curr_slot, k, v)
        curr_slot += 1
    return new

# Type signature of a CLIDriver action that returns a config dictionary.
CLIActionConfigType = Callable[[HammerDriver, Callable[[str], None]], Optional[dict]]

# Type signature of a CLIDriver action that returns a raw string.
CLIActionStringType = Callable[[HammerDriver, Callable[[str], None]], Optional[str]]

# CLIDriver action types.
CLIActionType = Union[CLIActionConfigType, CLIActionStringType]


# We cannot use isinstance() with CLIActionConfigType directly:
# "Parameterized generics cannot be used with class or instance checks"

def is_config_action(func: CLIActionType) -> bool:
    """Return True if the given function is a CLIActionConfigType."""
    return check_function_type(func, [HammerDriver, cast(type, Callable[[str], None])], cast(type, Optional[dict])) is None


def is_string_action(func: CLIActionType) -> bool:
    """Return True if the given function is a CLIActionConfigType."""
    return check_function_type(func, [HammerDriver, cast(type, Callable[[str], None])], cast(type, Optional[str])) is None


def check_CLIActionType_type(func: CLIActionType) -> None:
    """
    Check that the given CLIActionType obeys its function type signature.
    Raises TypeError if the function is of the incorrect type.
    """
    s = [HammerDriver, cast(type, Callable[[str], None])]
    config_check = check_function_type(func, s, cast(type, Optional[dict]))
    if config_check is None:
        return

    string_check = check_function_type(func, s, cast(type, Optional[str]))
    if string_check is None:
        return

    raise TypeError(
        "func does not appear to be a CLIActionType. Check for config returned {config}; check for string returned {string}".format(
            config=config_check, string=string_check))


class CLIDriver:
    """
    Helper class for projects to easily write/customize a CLI driver for hammer without needing to rewrite/copy all the
    argparse and plumbing.
    """

    def __init__(self) -> None:
        # Cache for the synthesis dir and par dir.
        # Defaults to blank (obj_dir + syn-rundir/par-rundir)
        self.syn_rundir = ""  # type: Optional[str]
        self.par_rundir = ""  # type: Optional[str]
        self.drc_rundir = ""  # type: Optional[str]
        self.lvs_rundir = ""  # type: Optional[str]
        self.sram_generator_rundir = ""  # type: Optional[str]
        self.sim_rundir = ""  # type: Optional[str]
        self.power_rundir = "" # type: Optional[str]
        self.formal_rundir = "" # type: Optional[str]
        self.timing_rundir = "" # type: Optional[str]
        self.pcb_rundir = ""  # type: Optional[str]

        # If a subclass has defined these, don't clobber them in init
        # since the subclass still uses this init function.
        if hasattr(self, "sram_generator_action"):
            check_CLIActionType_type(self.sram_generator_action)
        else:
            self.sram_generator_action = self.create_sram_generator_action([])  # type: CLIActionConfigType
        if hasattr(self, "pcb_action"):
            check_CLIActionType_type(self.pcb_action)
        else:
            self.pcb_action = self.create_pcb_action([])  # type: CLIActionConfigType
        if hasattr(self, "synthesis_action"):
            check_CLIActionType_type(self.synthesis_action)
        else:
            self.synthesis_action = self.create_synthesis_action([])  # type: CLIActionConfigType
        if hasattr(self, "par_action"):
            check_CLIActionType_type(self.par_action)
        else:
            self.par_action = self.create_par_action([])  # type: CLIActionConfigType
        if hasattr(self, "drc_action"):
            check_CLIActionType_type(self.drc_action)
        else:
            self.drc_action = self.create_drc_action([])  # type: CLIActionConfigType
        if hasattr(self, "lvs_action"):
            check_CLIActionType_type(self.lvs_action)
        else:
            self.lvs_action = self.create_lvs_action([])  # type: CLIActionConfigType
        if hasattr(self, "synthesis_par_action"):
            check_CLIActionType_type(self.synthesis_par_action)
        else:
            self.synthesis_par_action = self.create_synthesis_par_action(self.synthesis_action,
                                                                         self.par_action)  # type: CLIActionConfigType
        if hasattr(self, "sim_action"):
            check_CLIActionType_type(self.sim_action)
        else:
            self.sim_action = self.create_sim_action([])  # type: CLIActionConfigType
        if hasattr(self, "synthesis_sim_action"):
            check_CLIActionType_type(self.synthesis_sim_action)
        else:
            self.synthesis_sim_action = self.create_synthesis_sim_action(self.synthesis_action, self.sim_action)  # type: CLIActionConfigType
        if hasattr(self, "par_sim_action"):
            check_CLIActionType_type(self.par_sim_action)
        else:
            self.par_sim_action = self.create_par_sim_action(self.par_action, self.sim_action) # type: CLIActionConfigType
        if hasattr(self, "power_action"):
            check_CLIActionType_type(self.power_action)
        else:
            self.power_action = self.create_power_action([]) # type: CLIActionConfigType
        if hasattr(self, "formal_action"):
            check_CLIActionType_type(self.formal_action)
        else:
            self.formal_action = self.create_formal_action([]) # type: CLIActionConfigType
        if hasattr(self, "timing_action"):
            check_CLIActionType_type(self.timing_action)
        else:
            self.timing_action = self.create_timing_action([]) # type: CLIActionConfigType

        # Dictionaries of module-CLIActionConfigType for hierarchical flows.
        # See all_hierarchical_actions() below.
        self.hierarchical_synthesis_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_par_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_synthesis_par_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_drc_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_lvs_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_sim_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_power_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_formal_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_timing_actions = {}  # type: Dict[str, CLIActionConfigType]
        self.hierarchical_auto_action = None  # type: Optional[CLIActionConfigType]

    def action_map(self) -> Dict[str, CLIActionType]:
        """Return the mapping of valid actions -> functions for each action of the command-line driver."""
        return add_dicts({
            "build": self.generate_build_inputs,
            "build-inputs": self.generate_build_inputs,
            "build_inputs": self.generate_build_inputs,
            "dump": self.dump_action,
            "dump-macrosizes": self.dump_macrosizes_action,
            "dump_macrosizes": self.dump_macrosizes_action,
            "info": self.info_action,
            "sram-generator": self.sram_generator_action,
            "sram_generator": self.sram_generator_action,
            "pcb": self.pcb_action,
            "synthesis": self.synthesis_action,
            "syn": self.synthesis_action,
            "par": self.par_action,
            "synthesis_to_par": self.synthesis_to_par_action,
            "synthesis-to-par": self.synthesis_to_par_action,
            "syn_to_par": self.synthesis_to_par_action,
            "syn-to-par": self.synthesis_to_par_action,
            "synthesis_par": self.synthesis_par_action,
            "synthesis-par": self.synthesis_par_action,
            "syn_par": self.synthesis_par_action,
            "syn-par": self.synthesis_par_action,
            "hier_par_to_syn": self.hier_par_to_syn_action,
            "hier-par-to-syn": self.hier_par_to_syn_action,
            "par_to_drc": self.par_to_drc_action,
            "par-to-drc": self.par_to_drc_action,
            "par_to_lvs": self.par_to_lvs_action,
            "par-to-lvs": self.par_to_lvs_action,
            "drc": self.drc_action,
            "lvs": self.lvs_action,
            "sim": self.sim_action,
            "simulation": self.sim_action,
            "synthesis_to_sim": self.synthesis_to_sim_action,
            "synthesis-to-sim": self.synthesis_to_sim_action,
            "syn_to_sim": self.synthesis_to_sim_action,
            "syn-to-sim": self.synthesis_to_sim_action,
            "synthesis_sim": self.synthesis_sim_action,
            "synthesis-sim": self.synthesis_sim_action,
            "syn_sim": self.synthesis_sim_action,
            "syn-sim": self.synthesis_sim_action,
            "par_to_sim": self.par_to_sim_action,
            "par-to-sim": self.par_to_sim_action,
            "par_sim": self.par_sim_action,
            "par-sim": self.par_sim_action,
            "power": self.power_action,
            "syn-to-power": self.syn_to_power_action,
            "syn_to_power": self.syn_to_power_action,
            "par-to-power": self.par_to_power_action,
            "par_to_power": self.par_to_power_action,
            "sim-to-power": self.sim_to_power_action,
            "sim_to_power": self.sim_to_power_action,
            "formal": self.formal_action,
            "synthesis_to_formal": self.synthesis_to_formal_action,
            "synthesis-to-formal": self.synthesis_to_formal_action,
            "syn_to_formal": self.synthesis_to_formal_action,
            "syn-to-formal": self.synthesis_to_formal_action,
            "par_to_formal": self.par_to_formal_action,
            "par-to-formal": self.par_to_formal_action,
            "timing": self.timing_action,
            "synthesis_to_timing": self.synthesis_to_timing_action,
            "synthesis-to-timing": self.synthesis_to_timing_action,
            "syn_to_timing": self.synthesis_to_timing_action,
            "syn-to-timing": self.synthesis_to_timing_action,
            "par_to_timing": self.par_to_timing_action,
            "par-to-timing": self.par_to_timing_action
        }, self.all_hierarchical_actions)

    @staticmethod
    def dump_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """
        Just dump the parsed project configuration as the output.
        """
        return driver.project_config

    @staticmethod
    def info_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """
        Return the descripion of any key.
        """
        defaults_txt = importlib.resources.files("hammer.config").joinpath("defaults.yml").read_text()
        yaml = ruamel.yaml.YAML()
        data = yaml.load(defaults_txt)
        with open(KEY_PATH, 'r', encoding="utf-8") as f:
            yaml = ruamel.yaml.YAML()
            history = yaml.load(f)
        while True:
            curr_level = data
            overall_key = []
            while isinstance(curr_level, ruamel.yaml.CommentedMap):
                print()
                for k in curr_level.keys():
                    if "_meta" not in k:
                        print(k)
                while True:
                    try:
                        key = input("Select from the current level of keys: ")
                        next_level = curr_level[key]
                        break
                    except KeyError:
                        driver.log.error(f"Key {key} could not be found at the current level, try again.")
                overall_key.append(key)
                if not isinstance(next_level, ruamel.yaml.CommentedMap):
                    flat_key = '.'.join(overall_key)
                    if not driver.database.has_setting(flat_key):
                        val = curr_level.get(key)
                        driver.log.warning(f"{flat_key} is not in the project configuration, the default value is displayed.")
                    else:
                        val = driver.database.get_setting(flat_key)
                    if key in curr_level.ca.items:
                        comment = curr_level.ca.items[key][2].value.strip().replace('\n', ' ')  # NOTE: only takes in comment immediately after the key-value pair
                    else:
                        comment = "no comment provided"
                    key_hist = history[flat_key] if flat_key in history else "no history provided"
                    print(dedent(f"""
                    ----------------------------------------
                    Key: {flat_key}
                    Value: {val}
                    Description: {comment}
                    History: {key_hist}
                    ----------------------------------------
                    """))
                    while True:
                        continue_input = input("Continue querying keys? [y/n]: ")
                        if continue_input.lower() == 'y':
                            break
                        if continue_input.lower() == 'n':
                            return driver.project_config
                        driver.log.error("Please input either [y]es or [n]o.")
                    break
                curr_level = next_level

    @staticmethod
    def dump_macrosizes_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[str]:
        """
        Dump macro size information.
        """
        assert driver.tech is not None, "must have a technology"
        macro_json = list(map(lambda m: m.to_setting(), driver.tech.get_macro_sizes()))
        return json.dumps(macro_json, cls=HammerJSONEncoder, indent=4)

    def get_extra_synthesis_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra synthesis hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_par_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra place and route hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_drc_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra design-rule-check hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_lvs_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra layout-vs-schematic hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_sram_generator_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra SRAM generation hooks in this project.
        To be overriden by subclasses.
        """
        return list()

    def get_extra_sim_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra simulation hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_power_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra power hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_formal_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra formal hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_timing_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra timing hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def get_extra_pcb_hooks(self) -> List[HammerToolHookAction]:
        """
        Return a list of extra PCB deliverable hooks in this project.
        To be overridden by subclasses.
        """
        return list()

    def create_synthesis_action(self, custom_hooks: List[HammerToolHookAction],
                                pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                                post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                                post_run_func: Optional[Callable[[HammerDriver], None]] = None
                                ) -> CLIActionConfigType:
        hooks = self.get_extra_synthesis_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("synthesis", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_par_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_par_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("par", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    @staticmethod
    def get_full_config(driver: HammerDriver, output: dict) -> dict:
        """
        Get the full configuration by combining the project config from the
        driver with the given output dict (i.e. it contains only
        "synthesis.output.blah") that we want to combine with the project
        config.
        :param driver: HammerDriver that has the full project config.
        :param output: Output dict containing specific settings we want to add
                       to the full project config.
        :return: Full project config combined with the output dict
        """
        if "vlsi.builtins.is_complete" in output:
            if bool(output["vlsi.builtins.is_complete"]):
                raise ValueError("Output-only config claims it is complete")
        else:
            raise ValueError("Output-only config does not appear to be output only")

        output_full = deepdict(driver.project_config)
        output_full.update(deepdict(output))
        # Merged configs are always complete
        if "vlsi.builtins.is_complete" in output_full:
            del output_full["vlsi.builtins.is_complete"]
        return output_full

    def create_drc_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_drc_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("drc", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_lvs_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_lvs_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("lvs", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_sim_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_sim_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("sim", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_power_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_power_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("power", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_formal_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_formal_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("formal", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_timing_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_timing_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("timing", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_sram_generator_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_sram_generator_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("sram_generator", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_pcb_action(self, custom_hooks: List[HammerToolHookAction],
                          pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                          post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        hooks = self.get_extra_pcb_hooks() + custom_hooks  # type: List[HammerToolHookAction]
        return self.create_action("pcb", hooks if len(hooks) > 0 else None,
                                  pre_action_func, post_load_func, post_run_func)

    def create_action(self, action_type: str,
                      extra_hooks: Optional[List[HammerToolHookAction]],
                      pre_action_func: Optional[Callable[[HammerDriver], None]] = None,
                      post_load_func: Optional[Callable[[HammerDriver], None]] = None,
                      post_run_func: Optional[Callable[[HammerDriver], None]] = None) -> CLIActionConfigType:
        """
        Create an action function for the action_map.

        :param action_type: Either "syn"/"synthesis" or "par"
        :param extra_hooks: List of hooks to pass to the run function.
        :param pre_action_func: Optional function to call before doing anything.
        :param post_load_func: Optional function to call after loading the tool.
        :param post_run_func: Optional function to call after running the tool.
        :return: Action function.
        """

        def post_load_func_checked(driver: HammerDriver) -> None:
            """Check that post_load_func isn't null before calling it."""
            if post_load_func is not None:
                post_load_func(driver)

        def post_run_func_checked(driver: HammerDriver) -> None:
            """Check that post_run_func isn't null before calling it."""
            if post_run_func is not None:
                post_run_func(driver)

        def action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
            if pre_action_func is not None:
                pre_action_func(driver)

            # If the driver didn't successfully load, return None.
            # Note the step/hook priority is (in order of lowest to highest):
            # 1. Tool-supplied core steps
            # 2. Tool-supplied hooks (usually for persistence)
            # 3. Tech-supplied hooks
            # 4. User-supplied hooks
            assert driver.tech is not None, "must have a technology"
            with open(KEY_PATH, 'r') as f:
                key_history = json.load(f)
            if action_type == "synthesis" or action_type == "syn":
                if not driver.load_synthesis_tool(get_or_else(self.syn_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.syn_tool is not None, "load_synthesis_tool was unsuccessful"
                success, output = driver.run_synthesis(
                        driver.syn_tool.get_tool_hooks() + \
                        driver.tech.get_tech_syn_hooks(driver.syn_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("Synthesis tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.syn_tool.run_dir, "syn-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.syn_tool.run_dir, "syn-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.syn_tool.run_dir, "syn-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "par":
                if not driver.load_par_tool(get_or_else(self.par_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.par_tool is not None, "load_par_tool was unsuccessful"
                success, output = driver.run_par(
                        driver.par_tool.get_tool_hooks() + \
                        driver.tech.get_tech_par_hooks(driver.par_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("Place-and-route tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.par_tool.run_dir, "par-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.par_tool.run_dir, "par-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.par_tool.run_dir, "par-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "drc":
                if not driver.load_drc_tool(get_or_else(self.drc_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.drc_tool is not None, "load_drc_tool was unsuccessful"
                success, output = driver.run_drc(
                        driver.drc_tool.get_tool_hooks() + \
                        driver.tech.get_tech_drc_hooks(driver.drc_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("DRC tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.drc_tool.run_dir, "drc-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.drc_tool.run_dir, "drc-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.drc_tool.run_dir, "drc-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "lvs":
                if not driver.load_lvs_tool(get_or_else(self.lvs_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.lvs_tool is not None, "load_lvs_tool was unsuccessful"
                success, output = driver.run_lvs(
                        driver.lvs_tool.get_tool_hooks() + \
                        driver.tech.get_tech_lvs_hooks(driver.lvs_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("LVS tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.lvs_tool.run_dir, "lvs-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.lvs_tool.run_dir, "lvs-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.lvs_tool.run_dir, "lvs-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "sram_generator":
                if not driver.load_sram_generator_tool(get_or_else(self.sram_generator_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.sram_generator_tool is not None, "load_sram_generator_tool was unsuccessful"
                success, output = driver.run_sram_generator(
                        driver.sram_generator_tool.get_tool_hooks() + \
                        driver.tech.get_tech_sram_generator_hooks(driver.sram_generator_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("SRAM generator tool did not succeed")
                    return None
                post_run_func_checked(driver)
            elif action_type == "sim":
                if not driver.load_sim_tool(get_or_else(self.sim_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.sim_tool is not None, "load_sim_tool was unsuccessful"
                success, output = driver.run_sim(
                        driver.sim_tool.get_tool_hooks() + \
                        driver.tech.get_tech_sim_hooks(driver.sim_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("Sim tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.sim_tool.run_dir, "sim-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.sim_tool.run_dir, "sim-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.sim_tool.run_dir, "sim-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "power":
                if not driver.load_power_tool(get_or_else(self.power_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.power_tool is not None, "load_power_tool was unsuccessful"
                success, output = driver.run_power(extra_hooks)
                if not success:
                    driver.log.error("Power tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.power_tool.run_dir, "power-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.power_tool.run_dir, "power-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.power_tool.run_dir, "power-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "formal":
                if not driver.load_formal_tool(get_or_else(self.formal_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.formal_tool is not None, "load_formal_tool was unsuccessful"
                success, output = driver.run_formal(
                        driver.formal_tool.get_tool_hooks() + \
                        driver.tech.get_tech_formal_hooks(driver.formal_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("Formal tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.formal_tool.run_dir, "formal-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.formal_tool.run_dir, "formal-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.formal_tool.run_dir, "formal-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            elif action_type == "timing":
                if not driver.load_timing_tool(get_or_else(self.timing_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.timing_tool is not None, "load_timing_tool was unsuccessful"
                success, output = driver.run_timing(
                        driver.timing_tool.get_tool_hooks() + \
                        driver.tech.get_tech_timing_hooks(driver.timing_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("Timing tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.timing_tool.run_dir, "timing-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.timing_tool.run_dir, "timing-output-full.json"),
                                         self.get_full_config(driver, output))
                post_run_func_checked(driver)
            elif action_type == "pcb":
                if not driver.load_pcb_tool(get_or_else(self.pcb_rundir, "")):
                    return None
                else:
                    post_load_func_checked(driver)
                assert driver.pcb_tool is not None, "load_pcb_tool was unsuccessful"
                success, output = driver.run_pcb(
                        driver.pcb_tool.get_tool_hooks() + \
                        driver.tech.get_tech_pcb_hooks(driver.pcb_tool.name) + \
                        list(extra_hooks or []))
                if not success:
                    driver.log.error("PCB deliverable tool did not succeed")
                    return None
                dump_config_to_json_file(os.path.join(driver.pcb_tool.run_dir, "pcb-output.json"), output)
                dump_config_to_json_file(os.path.join(driver.pcb_tool.run_dir, "pcb-output-full.json"),
                                         self.get_full_config(driver, output))
                if driver.dump_history:
                    dump_config_to_yaml_file(os.path.join(driver.pcb_tool.run_dir, "pcb-output-history.yml"),
                                            add_key_history(self.get_full_config(driver, output), key_history))
                post_run_func_checked(driver)
            else:
                raise ValueError("Invalid action_type = " + str(action_type))
            # TODO: detect errors
            return output

        return action

    def synthesis_to_par_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        par_input_only = HammerDriver.synthesis_output_to_par_input(driver.project_config)
        if par_input_only is None:
            driver.log.error("Input config does not appear to contain valid synthesis outputs")
            return None
        else:
            return self.get_full_config(driver, par_input_only)

    def synthesis_to_sim_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        sim_input_only = HammerDriver.synthesis_output_to_sim_input(driver.project_config)
        if sim_input_only is None:
            driver.log.error("Input config does not appear to contain valid synthesis outputs")
            return None
        else:
            return self.get_full_config(driver, sim_input_only)

    def par_to_sim_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        sim_input_only = HammerDriver.par_output_to_sim_input(driver.project_config)
        if sim_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, sim_input_only)

    def hier_par_to_syn_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """ Create a full config to run the output. """
        syn_input_only = HammerDriver.par_output_to_syn_input(driver.project_config)
        if syn_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, syn_input_only)

    def par_to_drc_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """ Create a full config to run the output. """
        drc_input_only = HammerDriver.par_output_to_drc_input(driver.project_config)
        if drc_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, drc_input_only)

    def par_to_lvs_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """ Create a full config to run the output. """
        lvs_input_only = HammerDriver.par_output_to_lvs_input(driver.project_config)
        if lvs_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, lvs_input_only)

    def syn_to_power_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        power_input_only = HammerDriver.synthesis_output_to_power_input(driver.project_config)
        if power_input_only is None:
            driver.log.error("Input config does not appear to contain valid syn outputs")
            return None
        else:
            return self.get_full_config(driver, power_input_only)

    def par_to_power_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        power_input_only = HammerDriver.par_output_to_power_input(driver.project_config)
        if power_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, power_input_only)

    def sim_to_power_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        power_input_only = HammerDriver.sim_output_to_power_input(driver.project_config)
        if power_input_only is None:
            driver.log.error("Input config does not appear to contain valid sim outputs")
            return None
        else:
            return self.get_full_config(driver, power_input_only)

    def synthesis_to_formal_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        formal_input_only = HammerDriver.synthesis_output_to_formal_input(driver.project_config)
        if formal_input_only is None:
            driver.log.error("Input config does not appear to contain valid synthesis outputs")
            return None
        else:
            return self.get_full_config(driver, formal_input_only)

    def par_to_formal_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        formal_input_only = HammerDriver.par_output_to_formal_input(driver.project_config)
        if formal_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, formal_input_only)

    def synthesis_to_timing_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        timing_input_only = HammerDriver.synthesis_output_to_timing_input(driver.project_config)
        if timing_input_only is None:
            driver.log.error("Input config does not appear to contain valid synthesis outputs")
            return None
        else:
            return self.get_full_config(driver, timing_input_only)

    def par_to_timing_action(self, driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """Create a full config to run the output."""
        timing_input_only = HammerDriver.par_output_to_timing_input(driver.project_config)
        if timing_input_only is None:
            driver.log.error("Input config does not appear to contain valid par outputs")
            return None
        else:
            return self.get_full_config(driver, timing_input_only)


    def create_synthesis_par_action(self, synthesis_action: CLIActionConfigType, par_action: CLIActionConfigType) -> CLIActionConfigType:
        """
        Create a parameterizable synthesis_par action for the CLIDriver.

        :param synthesis_action: synthesis action
        :param par_action: par action
        :return: Custom synthesis_par action
        """

        def syn_par_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
            # Synthesis output.
            syn_output = synthesis_action(driver, append_error_func)
            if syn_output is None:
                append_error_func("Synthesis action in syn_par failed")
                return None
            else:
                # Generate place-and-route input from the synthesis output.
                syn_output_converted = HammerDriver.synthesis_output_to_par_input(syn_output)
                assert syn_output_converted is not None, "syn_output must be generated by CLIDriver"
                par_input = self.get_full_config(driver, syn_output_converted)  # type: dict

                # Dump both synthesis output and par input for debugging/resuming.
                # TODO(edwardw): make these output filenames configurable?
                assert driver.syn_tool is not None, "Syn tool must exist since we ran synthesis_action successfully"
                dump_config_to_json_file(os.path.join(driver.syn_tool.run_dir, "par-input.json"), par_input)

                # Use new par input and run place-and-route.
                driver.update_project_configs([par_input])
                par_output = par_action(driver, append_error_func)
                return par_output

        return syn_par_action

    def create_synthesis_sim_action(self, synthesis_action: CLIActionConfigType, sim_action: CLIActionConfigType) -> CLIActionConfigType:
        """
        Create a parameterizable synthesis_sim action for the CLIDriver.

        :param synthesis_action: synthesis action
        :param sim_action: sim action
        :return: Custom synthesis_sim action
        """

        def syn_sim_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
            # Synthesis output.
            syn_output = synthesis_action(driver, append_error_func)
            if syn_output is None:
                append_error_func("Synthesis action in syn_sim failed")
                return None
            else:
                # Generate sim input from the synthesis output.
                syn_output_converted = HammerDriver.synthesis_output_to_sim_input(syn_output)
                assert syn_output_converted is not None, "syn_output must be generated by CLIDriver"
                sim_input = self.get_full_config(driver, syn_output_converted)  # type: dict

                # Dump both synthesis output and sim input for debugging/resuming.
                assert driver.syn_tool is not None, "Syn tool must exist since we ran synthesis_action successfully"
                dump_config_to_json_file(os.path.join(driver.syn_tool.run_dir, "sim-input.json"), sim_input)

                # Use new sim input and run simulation.
                driver.update_project_configs([sim_input])
                sim_output = sim_action(driver, append_error_func)
                return sim_output

        return syn_sim_action

    def create_par_sim_action(self, par_action: CLIActionConfigType, sim_action: CLIActionConfigType) -> CLIActionConfigType:
        """
        Create a parameterizable par_sim action for the CLIDriver.

        :param par_action: par action
        :param sim_action: sim action
        :return: Custom par_sim action
        """

        def par_sim_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
            # Synthesis output.
            par_output = par_action(driver, append_error_func)
            if par_output is None:
                append_error_func("PAR action in syn_sim failed")
                return None
            else:
                # Generate sim input from the par output.
                par_output_converted = HammerDriver.par_output_to_sim_input(par_output)
                assert par_output_converted is not None, "par_output must be generated by CLIDriver"
                sim_input = self.get_full_config(driver, par_output_converted)  # type: dict

                # Dump both par output and sim input for debugging/resuming.
                assert driver.par_tool is not None, "PAR tool must exist since we ran par_action successfully"
                dump_config_to_json_file(os.path.join(driver.par_tool.run_dir, "sim-input.json"), sim_input)

                # Use new sim input and run simulation.
                driver.update_project_configs([sim_input])
                sim_output = sim_action(driver, append_error_func)
                return sim_output

        return par_sim_action

    ### Hierarchical stuff ###
    @property
    def all_hierarchical_actions(self) -> Dict[str, CLIActionConfigType]:
        """
        Return a list of hierarchical actions if the given project configuration is a hierarchical design.
        Set when the driver is first created in args_to_driver.
        Create syn/synthesis-[block], par-[block], and /syn_par-[block].

        :return: Dictionary of actions to use (could be empty).
        """
        actions = {}  # type: Dict[str, CLIActionConfigType]
        if self.hierarchical_auto_action is not None:
            actions.update({"auto": self.hierarchical_auto_action})

        def add_variants(templates: List[str], block: str, action: CLIActionConfigType) -> None:
            """Just add the given action using the name templates."""
            for template in templates:
                name = template.format(block=block)
                actions.update({name: action})

        for module, action in self.hierarchical_synthesis_actions.items():
            add_variants([
                "syn-{block}",
                "synthesis-{block}",
                "syn_{block}",
                "synthesis_{block}"
            ], module, action)

        for module, action in self.hierarchical_par_actions.items():
            add_variants([
                "par-{block}",
                "par_{block}"
            ], module, action)

        for module, action in self.hierarchical_synthesis_par_actions.items():
            add_variants([
                "syn-par-{block}",
                "syn_par-{block}",
                "syn-par_{block}",
                "syn_par_{block}"
            ], module, action)

        for module, action in self.hierarchical_drc_actions.items():
            add_variants([
                "drc-{block}",
                "drc_{block}"
            ], module, action)

        for module, action in self.hierarchical_lvs_actions.items():
            add_variants([
                "lvs-{block}",
                "lvs_{block}"
            ], module, action)

        for module, action in self.hierarchical_sim_actions.items():
            add_variants([
                "sim-{block}",
                "sim_{block}"
            ], module, action)

        for module, action in self.hierarchical_power_actions.items():
            add_variants([
                "power-{block}",
                "power_{block}"
            ], module, action)

        for module, action in self.hierarchical_formal_actions.items():
            add_variants([
                "formal-{block}",
                "formal_{block}"
            ], module, action)

        for module, action in self.hierarchical_timing_actions.items():
            add_variants([
                "timing-{block}",
                "timing_{block}"
            ], module, action)

        return actions

    def get_extra_hierarchical_synthesis_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical synthesis hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_par_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical place and route hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_drc_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical DRC hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_lvs_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical LVS hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_sim_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical sim hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_power_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical power hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_formal_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical formal hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    def get_extra_hierarchical_timing_hooks(self, driver: HammerDriver) -> Dict[str, List[HammerToolHookAction]]:
        """
        Return a list of extra hierarchical timing hooks in this project.
        To be overridden by subclasses.

        :return: Dictionary of (module name, list of hooks)
        """
        return dict()

    # The following functions are present for further user customizability.

    def get_hierarchical_synthesis_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical synthesis for the given module (in hierarchical flows).
        """
        return self.hierarchical_synthesis_actions[module]

    def set_hierarchical_synthesis_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical synthesis for the given module (in hierarchical flows).
        """
        self.hierarchical_synthesis_actions[module] = action

    def get_hierarchical_par_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical par for the given module (in hierarchical flows).
        """
        return self.hierarchical_par_actions[module]

    def set_hierarchical_par_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical par for the given module (in hierarchical flows).
        """
        self.hierarchical_par_actions[module] = action

    def set_hierarchical_drc_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical drc for the given module (in hierarchical flows).
        """
        self.hierarchical_drc_actions[module] = action

    def get_hierarchical_drc_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical drc for the given module (in hierarchical flows).
        """
        return self.hierarchical_drc_actions[module]

    def set_hierarchical_lvs_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical lvs for the given module (in hierarchical flows).
        """
        self.hierarchical_lvs_actions[module] = action

    def get_hierarchical_lvs_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical lvs for the given module (in hierarchical flows).
        """
        return self.hierarchical_lvs_actions[module]

    def get_hierarchical_synthesis_par_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical syn_par for the given module (in hierarchical flows).
        """
        return self.hierarchical_synthesis_par_actions[module]

    def set_hierarchical_synthesis_par_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical syn_par for the given module (in hierarchical flows).
        """
        self.hierarchical_synthesis_par_actions[module] = action

    def set_hierarchical_sim_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical sim for the given module (in hierarchical flows).
        """
        self.hierarchical_sim_actions[module] = action

    def get_hierarchical_sim_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical sim for the given module (in hierarchical flows).
        """
        return self.hierarchical_sim_actions[module]

    def set_hierarchical_power_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical power for the given module (in hierarchical flows).
        """
        self.hierarchical_power_actions[module] = action

    def get_hierarchical_power_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical power for the given module (in hierarchical flows).
        """
        return self.hierarchical_power_actions[module]

    def set_hierarchical_formal_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical formal for the given module (in hierarchical flows).
        """
        self.hierarchical_formal_actions[module] = action

    def get_hierarchical_formal_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical formal for the given module (in hierarchical flows).
        """
        return self.hierarchical_formal_actions[module]

    def set_hierarchical_timing_action(self, module: str, action: CLIActionConfigType) -> None:
        """
        Set the action associated with hierarchical timing for the given module (in hierarchical flows).
        """
        self.hierarchical_timing_actions[module] = action

    def get_hierarchical_timing_action(self, module: str) -> CLIActionConfigType:
        """
        Get the action associated with hierarchical timing for the given module (in hierarchical flows).
        """
        return self.hierarchical_timing_actions[module]

    def valid_actions(self) -> List[str]:
        """Get the list of valid actions for the command-line driver."""
        return list(self.action_map().keys())

    def args_to_driver(self, args: dict,
                       default_options: Optional[HammerDriverOptions] = None) -> \
            Tuple[HammerDriver, List[str]]:
        """Parse command line arguments and environment variables for the command line front-end to hammer-vlsi.

        :return: HammerDriver and a list of errors."""

        # TODO: rewrite this less tediously?

        # Resolve default_options.
        # Can't call HammerDriver.get_default_driver_options in the
        # parameters as it will be called when args_to_driver is defined
        default_options_resolved = HammerDriver.get_default_driver_options()  # type: HammerDriverOptions
        if default_options is not None:
            default_options_resolved = default_options

        # Driver options.
        options = default_options_resolved  # type: HammerDriverOptions

        # Extra config (flattened JSON).
        config = {}  # type: Dict[str, Any]

        # Create a list of errors for the user.
        errors = []  # type: List[str]

        # Load environment configs.
        env_configs = parse_optional_file_list_from_args(args['environment_config'],
                                                         append_error_func=errors.append)  # type: List[str]
        # Also load any environment configs from the environment.
        split_env_var_s = os.environ.get("HAMMER_ENVIRONMENT_CONFIGS", default="").split(os.pathsep)  # type: List[str]
        # "".split(':') returns [''], so we need to catch this case and return an empty list as intended.
        split_env_var = [] if split_env_var_s == [''] else split_env_var_s  # type: List[str]
        env_config_environment_var = parse_optional_file_list_from_args(split_env_var,
                                                                        append_error_func=errors.append)  # type: List[str]
        for extra_path in env_config_environment_var:
            env_configs.append(extra_path)
        options = options._replace(environment_configs=list(env_configs))

        # Load project configs.
        project_configs = parse_optional_file_list_from_args(args['configs'], append_error_func=errors.append)
        options = options._replace(project_configs=list(project_configs))

        # Log file.
        log = args["log"]
        if log is not None:
            if isinstance(log, str):
                options = options._replace(log_file=log)
            else:
                errors.append("Log file 'log' is not a string")

        # Verilog inputs.
        # (optional, since it can also be specified from JSON)
        verilogs = args['verilog']
        if isinstance(verilogs, List) and len(verilogs) > 0:
            config.update({'synthesis.inputs.input_files': list(verilogs)})

        # Top module.
        # (optional, since it can also be specified from JSON)
        top_module = get_nonempty_str(args['top'])
        if top_module is not None:
            config['synthesis.inputs.top_module'] = top_module

        # Object dir.
        # (optional)
        obj_dir = get_nonempty_str(args['obj_dir'])
        if obj_dir is None:
            # Try getting object dir from environment variable.
            obj_dir = get_nonempty_str(os.environ.get("HAMMER_DRIVER_OBJ_DIR", ""))
        if obj_dir is not None:
            options = options._replace(obj_dir=os.path.realpath(obj_dir))
        # Syn/par rundir (optional)
        self.syn_rundir = get_nonempty_str(args['syn_rundir'])
        self.par_rundir = get_nonempty_str(args['par_rundir'])
        self.drc_rundir = get_nonempty_str(args['drc_rundir'])
        self.lvs_rundir = get_nonempty_str(args['lvs_rundir'])
        self.sim_rundir = get_nonempty_str(args['sim_rundir'])
        self.power_rundir = get_nonempty_str(args['power_rundir'])
        self.formal_rundir = get_nonempty_str(args['formal_rundir'])
        self.timing_rundir = get_nonempty_str(args['timing_rundir'])

        # Intercept keys for determining key origin
        project_configs_yaml: List[dict] = []
        for conf_file in project_configs:
            config_str = Path(conf_file).read_text()
            project_configs_yaml.append(hammer.config.load_config_from_string(config_str, is_yaml=True, path=str(Path(conf_file).resolve().parent)))
        project_configs_yaml_keys = [set(i.keys()) for i in project_configs_yaml]
        key_history: Dict[str, List[str]] = {i: [] for i in reduce(lambda x, y: x.union(y), project_configs_yaml_keys)}
        for cfg_file, cfg in zip(project_configs, project_configs_yaml_keys):
            for key in cfg:
                key_history[key].append(cfg_file)
        with open(KEY_PATH, 'w') as f:
            json.dump(key_history, f)

        # Stage control: from/to
        from_step = get_nonempty_str(args['from_step'])
        after_step = get_nonempty_str(args['after_step'])
        to_step = get_nonempty_str(args['to_step'])
        until_step = get_nonempty_str(args['until_step'])
        only_step = get_nonempty_str(args['only_step'])

        if from_step is not None and after_step is not None:
            errors.append("Specified both start_before_step and start_after_step. start_before_step will take precedence.")
        if to_step is not None and until_step is not None:
            errors.append("Specified both stop_after_step and stop_before_step. stop_after_step will take precedence.")
        if only_step is not None and any(s is not None for s in [from_step, after_step, to_step, until_step]):
            errors.append("Specified {start|stop}_{before_after}_step with only_step. only_step will take precedence.")
        if (from_step is not None and until_step is not None and from_step == until_step) or \
           (after_step is not None and to_step is not None and after_step == to_step) or \
           (after_step is not None and until_step is not None and after_step == until_step):
            errors.append("Caution: start_before_step == stop_before_step, start_after_step == stop_after_step, or start_after_step == stop_before_step will result in nothing being run")

        driver = HammerDriver(options, config)

        start_step = only_step or from_step or after_step or None
        start_incl = (only_step or from_step) is not None
        stop_step = only_step or to_step or until_step or None
        stop_incl = (only_step or to_step) is not None
        if (start_step or stop_step) is not None:
            driver.set_post_custom_syn_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_par_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_drc_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_lvs_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_sim_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_power_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_formal_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))
            driver.set_post_custom_timing_tool_hooks(HammerTool.make_start_stop_hooks(
                HammerStartStopStep(step=start_step, inclusive=start_incl),
                HammerStartStopStep(step=stop_step, inclusive=stop_incl)))

        # Hierarchical support.
        # Generate synthesis and par actions for each module above.
        hierarchical_settings = driver.get_hierarchical_settings()

        for module_iter, config_iter in hierarchical_settings:
            def create_actions(module: str, config: dict) -> None:
                # Create a new context (this def) per module, otherwise when these higher-order funcs run they'll all
                # use the last iteration of the loop.

                # TODO(edwardw): this is a bit of a hack.
                # Should really add an API to allow a run to have a bit of temporary project config
                base_project_config = [[]]  # type: List[List[dict]]

                def syn_pre_func(d: HammerDriver) -> None:
                    self.syn_rundir = os.path.join(d.obj_dir, "syn-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def par_pre_func(d: HammerDriver) -> None:
                    self.par_rundir = os.path.join(d.obj_dir, "par-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def drc_pre_func(d: HammerDriver) -> None:
                    self.drc_rundir = os.path.join(d.obj_dir, "drc-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def lvs_pre_func(d: HammerDriver) -> None:
                    self.lvs_rundir = os.path.join(d.obj_dir, "lvs-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def sim_pre_func(d: HammerDriver) -> None:
                    self.lvs_rundir = os.path.join(d.obj_dir, "sim-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def power_pre_func(d: HammerDriver) -> None:
                    self.lvs_rundir = os.path.join(d.obj_dir, "power-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def formal_pre_func(d: HammerDriver) -> None:
                    self.lvs_rundir = os.path.join(d.obj_dir, "formal-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def timing_pre_func(d: HammerDriver) -> None:
                    self.lvs_rundir = os.path.join(d.obj_dir, "timing-{module}".format(
                        module=module))  # TODO(edwardw): fix this ugly os.path.join; it doesn't belong here.
                    # TODO(edwardw): remove ugly hack to store stuff in parent context
                    base_project_config[0] = deeplist(driver.project_configs)
                    d.update_project_configs(deeplist(base_project_config[0]) + [config])

                def post_run(d: HammerDriver, rundir: str) -> None:
                    # Write out the configs used/generated for logging/debugging.
                    with open(os.path.join(rundir, "full_config.json"), "w") as f:
                        new_output_json = json.dumps(d.project_config, cls=HammerJSONEncoder, indent=4)
                        f.write(new_output_json)
                    with open(os.path.join(rundir, "module_config.json"), "w") as f:
                        new_output_json = json.dumps(config, cls=HammerJSONEncoder, indent=4)
                        f.write(new_output_json)

                    d.update_project_configs(deeplist(base_project_config[0]))

                def syn_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.syn_rundir, ""))

                def par_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.par_rundir, ""))

                def drc_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.drc_rundir, ""))

                def lvs_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.lvs_rundir, ""))

                def sim_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.sim_rundir, ""))

                def power_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.power_rundir, ""))

                def formal_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.formal_rundir, ""))

                def timing_post_run(d: HammerDriver) -> None:
                    post_run(d, get_or_else(self.timing_rundir, ""))

                syn_action = self.create_synthesis_action(self.get_extra_hierarchical_synthesis_hooks(driver).get(module, []),
                                                          pre_action_func=syn_pre_func, post_load_func=None,
                                                          post_run_func=syn_post_run)
                self.set_hierarchical_synthesis_action(module, syn_action)
                par_action = self.create_par_action(self.get_extra_hierarchical_par_hooks(driver).get(module, []),
                                                    pre_action_func=par_pre_func, post_load_func=None,
                                                    post_run_func=par_post_run)
                self.set_hierarchical_par_action(module, par_action)
                syn_par_action = self.create_synthesis_par_action(synthesis_action=syn_action, par_action=par_action)
                self.set_hierarchical_synthesis_par_action(module, syn_par_action)
                drc_action = self.create_drc_action(self.get_extra_hierarchical_drc_hooks(driver).get(module, []),
                                                    pre_action_func=drc_pre_func, post_load_func=None,
                                                    post_run_func=drc_post_run)
                self.set_hierarchical_drc_action(module, drc_action)
                lvs_action = self.create_lvs_action(self.get_extra_hierarchical_lvs_hooks(driver).get(module, []),
                                                    pre_action_func=lvs_pre_func, post_load_func=None,
                                                    post_run_func=lvs_post_run)
                self.set_hierarchical_lvs_action(module, lvs_action)
                sim_action = self.create_sim_action(self.get_extra_hierarchical_sim_hooks(driver).get(module, []),
                                                    pre_action_func=sim_pre_func, post_load_func=None,
                                                    post_run_func=sim_post_run)
                self.set_hierarchical_sim_action(module, sim_action)
                power_action = self.create_power_action(self.get_extra_hierarchical_power_hooks(driver).get(module, []),
                                                    pre_action_func=power_pre_func, post_load_func=None,
                                                    post_run_func=power_post_run)
                self.set_hierarchical_power_action(module, power_action)
                formal_action = self.create_formal_action(self.get_extra_hierarchical_formal_hooks(driver).get(module, []),
                                                    pre_action_func=formal_pre_func, post_load_func=None,
                                                    post_run_func=formal_post_run)
                self.set_hierarchical_formal_action(module, formal_action)
                timing_action = self.create_timing_action(self.get_extra_hierarchical_timing_hooks(driver).get(module, []),
                                                    pre_action_func=timing_pre_func, post_load_func=None,
                                                    post_run_func=timing_post_run)
                self.set_hierarchical_timing_action(module, timing_action)

            create_actions(module_iter, config_iter)

        # If we are in hierarchical mode, also generate an auto that can run the whole flow.
        if len(hierarchical_settings) > 0:
            def auto_action(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
                log = driver.log.context("CLIDriver_auto")
                output = {}  # type: dict

                # Run syn_par for every module.
                for module, _ in hierarchical_settings:
                    syn_par_action = self.get_hierarchical_synthesis_par_action(module)
                    new_output = syn_par_action(driver, append_error_func)

                    if new_output is None:
                        log.error("Hierarchical syn-par run for module {module} failed".format(module=module))
                        return None
                    else:
                        log.info("Hierarchical syn-par run for module {module} finished".format(module=module))

                    b, ext = os.path.splitext(args["output"])
                    new_output_filename = "{base}-{module}{ext}".format(base=b, module=module, ext=ext)
                    with open(new_output_filename, "w") as f:
                        new_output_json = json.dumps(new_output, cls=HammerJSONEncoder, indent=4)
                        f.write(new_output_json)
                    log.info("Output JSON: " + str(new_output))

                    new_ilm = {
                        "vlsi.inputs.ilms": new_output["par.outputs.output_ilms"],
                        "vlsi.inputs.ilms_meta": "append"
                    }
                    new_ilm_filename = "{base}-{module}_ilm{ext}".format(base=b, module=module, ext=ext)
                    with open(new_ilm_filename, "w") as f:
                        json_content = json.dumps(new_ilm, cls=HammerJSONEncoder, indent=4)
                        f.write(json_content)
                    log.info("New input ILM JSON written to " + new_ilm_filename)
                    driver.update_project_configs(driver.project_configs + [new_ilm])
                return output

            self.hierarchical_auto_action = auto_action

        driver.dump_history = args["dump_history"]

        return driver, errors

    @staticmethod
    def generate_build_inputs(driver: HammerDriver, append_error_func: Callable[[str], None]) -> Optional[dict]:
        """
        Generate the build tool artifacts for this flow, specified by the "vlsi.core.build_system" key.
        The flow is the set of steps configured by the current HammerIR input.

        :param driver: The HammerDriver object which has parsed the configs specified by -p
        :param append_error_func: The function to use to append an error
        :return: The diplomacy graph
        """
        build_system = str(driver.database.get_setting("vlsi.core.build_system", "none"))
        if build_system in BuildSystems:
            return BuildSystems[build_system](driver, append_error_func)
        else:
            raise ValueError("Unsupported build system: {}".format(build_system))

    def run_main_parsed(self, args: dict) -> int:
        """
        Given a parsed dictionary of arguments, find and run the given action.

        :return: Return code (0 for success)
        """
        if args['firrtl'] is not None and len(args['firrtl']) > 0:
            print("firrtl convenience argument not yet implemented", file=sys.stderr)
            return 1

        driver, errors = self.args_to_driver(args)

        # Check for action after creating the driver (e.g. for custom actions like hierarchical actions).
        action = str(args['action'])  # type: str
        if action not in self.valid_actions():
            print("Invalid action {action}".format(action=action), file=sys.stderr)
            print("Valid actions are: {actions}".format(actions=", ".join(self.valid_actions())), file=sys.stderr)
            return 1

        action_func = self.action_map()[action]
        output_str = None  # type: Optional[str]
        if is_config_action(action_func):
            action_func = cast(CLIActionConfigType, action_func)
            output_config = action_func(driver, errors.append)  # type: Optional[dict]
            if output_config is not None:
                output_str = json.dumps(output_config, cls=HammerJSONEncoder, indent=4)
        elif is_string_action(action_func):
            action_func = cast(CLIActionStringType, action_func)
            output_str = action_func(driver, errors.append)
        else:
            raise NotImplementedError("Invalid action function")
        if output_str is None:
            print("Action {action} failed with errors".format(action=action), file=sys.stderr)
            for err in errors:
                print(err, file=sys.stderr)
            return 1
        else:
            with open(args["output"], "w") as f:
                f.write(output_str)
            print("Action {action} config output written to {file}".format(action=action, file=args["output"]))
            return 0


    def main(self, args: Optional[List[str]] = None) -> None:
        """
        Main function to call from your entry point script.
        Parses command line arguments.
        :param args: Custom command-line arguments.  If not given, sys.argv[1:] will be used.
        Example:
        >>> if __name__ == '__main__':
        >>>   CLIDriver().main()
        """
        parser = argparse.ArgumentParser()

        parser.add_argument('action', metavar='ACTION', type=str,  # choices=self.valid_actions() <- sadly incompatible w/custom actions
                            help='Action to perform with the command-line driver.')
        # Required arguments for (Python) hammer driver.
        parser.add_argument("-e", "--environment_config", action='append', required=False,
                            help="Environment config files (.yml or .json) - .json will take precendence over any .yml. These config files will not be re-emitted in the output json. Can also be specified as a colon-separated list in the environment variable HAMMER_ENVIRONMENT_CONFIGS.")
        parser.add_argument("-p", "--project_config", action='append', dest="configs", type=str,
                            help='Project config files (.yml or .json).')
        parser.add_argument("-l", "--log", required=False,
                            help='Log file. Leave blank to automatically create one.')
        parser.add_argument("--obj_dir", required=False,
                            help='Folder for storing results of CAD tool runs. If not specified, this will be the hammer-vlsi folder by default. Can also be specified via the environment variable HAMMER_DRIVER_OBJ_DIR.')
        parser.add_argument("--syn_rundir", required=False, default="",
                            help='(optional) Directory to store syn results in')
        parser.add_argument("--par_rundir", required=False, default="",
                            help='(optional) Directory to store par results in')
        parser.add_argument("--drc_rundir", required=False, default="",
                            help='(optional) Directory to store DRC results in')
        parser.add_argument("--lvs_rundir", required=False, default="",
                            help='(optional) Directory to store LVS results in')
        parser.add_argument("--sim_rundir", required=False, default="",
                            help='(optional) Directory to store simulation results in')
        parser.add_argument("--power_rundir", required=False, default="",
                            help='(optional) Directory to store power results in')
        parser.add_argument("--formal_rundir", required=False, default="",
                            help='(optional) Directory to store formal results in')
        parser.add_argument("--timing_rundir", required=False, default="",
                            help='(optional) Directory to store timing results in')
        # Optional arguments for step control.
        parser.add_argument("--start_before_step", "--from_step", dest="from_step", required=False,
                            help="Run the given action from before the given step (inclusive). Not compatible with --start_after_step.")
        parser.add_argument("--start_after_step", "--after_step", dest="after_step", required=False,
                            help="Run the given action from after the given step (exclusive). Not compatible with --start_before_step.")
        parser.add_argument("--stop_after_step", "--to_step", dest="to_step", required=False,
                            help="Run the given action to the given step (inclusive). Not compatible with --stop_before_step.")
        parser.add_argument("--stop_before_step", "--until_step", dest="until_step", required=False,
                            help="Run the given action until the given step (exclusive). Not compatible with --stop_after_step.")
        parser.add_argument("--only_step", dest="only_step", required=False,
                            help="Run only the given step. Not compatible with --{start|stop}_{before|after}_step.")
        # Required arguments for CLI hammer driver.
        parser.add_argument("-o", "--output", default="output.json", required=False,
                            help='Output JSON file for results and modular use of hammer-vlsi. Default: output.json.')
        # Optional arguments (depending on context)
        parser.add_argument("-v", "--verilog", required=False, action='append',
                            help='Input set of Verilog files.')
        parser.add_argument("-f", "--firrtl", action='append', required=False,
                            help='Input set of firrtl files. Provided for convenience; hammer-vlsi will transform it to Verilog.')
        parser.add_argument("-t", "--top", required=False,
                            help='Top module. If not specified, hammer-vlsi will take it from synthesis.inputs.top_module.')
        parser.add_argument("--cad-files", action='append', required=False,
                            help="CAD files.")
        parser.add_argument("--dump-history", default=False, action=argparse.BooleanOptionalAction,
                            help='Option to dump the key history of all the project configurations.')

        try:
            output = subprocess.check_output(["hammer-shell-test"]).decode().strip()
        except FileNotFoundError:
            output = "File not found"
        except PermissionError as e:
            output = str(e.args[0]) + " " + e.args[1]
        if output != "hammer-shell appears to be on the path":
            print("hammer-shell does not appear to be on the path (hammer-shell-test failed to run: %s)" % (output),
                  file=sys.stderr)
            sys.exit(1)

        sys.exit(self.run_main_parsed(vars(parser.parse_args(args))))



#  hammer-vlsi plugin for Cadence Joules.
#
#  See LICENSE for licence details.

from typing import List, Dict, Optional

import os

from hammer.vlsi import HammerPowerTool, HammerToolStep, MMMCCornerType, FlowLevel
from hammer.logging import HammerVLSILogging

from hammer.common.cadence import CadenceTool


class Joules(HammerPowerTool, CadenceTool):

    @property
    def post_synth_sdc(self) -> Optional[str]:
        return None

    def tool_config_prefix(self) -> str:
        return "power.joules"

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = dict(super().env_vars)
        new_dict["JOULES_BIN"] = self.get_setting("power.joules.joules_bin")
        return new_dict

    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.check_level,
            self.init_technology,
            self.init_design,
            self.read_stimulus,
            self.synthesize_design,
            self.compute_power,
            self.report_power,
            self.run_joules
        ])

    def check_level(self) -> bool:
        if self.level == FlowLevel.RTL or self.level == FlowLevel.SYN:
            return True
        else:
            self.logger.error("The FlowLevel is invalid. The Joules plugin only supports RTL and post-synthesis analysis. Check your power tool setting and flow step.")
            return False

    def init_technology(self) -> bool:
        # libs, define RAMs, define corners
        verbose_append = self.verbose_append
        verbose_append("set_multi_cpu_usage -local_cpu {}".format(self.get_setting("vlsi.core.max_threads")))

        corners = self.get_mmmc_corners()
        if MMMCCornerType.Extra in list(map(lambda corner: corner.type, corners)):
            for corner in corners:
                if corner.type is MMMCCornerType.Extra:
                    verbose_append("read_libs {EXTRA_LIBS} -domain extra -infer_memory_cells".format(EXTRA_LIBS=self.get_timing_libs(corner)))
                    break
        elif MMMCCornerType.Setup in list(map(lambda corner: corner.type, corners)):
            for corner in corners:
                if corner.type is MMMCCornerType.Setup:
                    verbose_append("read_libs {SETUP_LIBS} -domain setup -infer_memory_cells".format(SETUP_LIBS=self.get_timing_libs(corner)))
                    break
        elif MMMCCornerType.Hold in list(map(lambda corner: corner.type, corners)):
            for corner in corners:
                if corner.type is MMMCCornerType.Hold:
                    verbose_append("read_libs {HOLD_LIBS} -domain hold -infer_memory_cells".format(HOLD_LIBS=self.get_timing_libs(corner)))
                    break
        else:
            self.logger.error("No corners found")
            return False
        return True

    def init_design(self) -> bool:
        verbose_append = self.verbose_append

        top_module = self.get_setting("power.inputs.top_module")
        tb_name = self.tb_name
        # Replace . to / formatting in case argument passed from sim tool
        tb_dut = self.tb_dut.replace(".", "/")


        if self.level == FlowLevel.RTL:
            # Read in the design files
            verbose_append("read_hdl -sv {}".format(" ".join(self.input_files)))

        # Setup the power specification
        power_spec_arg = self.map_power_spec_name()
        power_spec_file = self.create_power_spec()
        if not power_spec_arg or not power_spec_file:
            return False

        verbose_append("read_power_intent -{tpe} {spec} -module {TOP_MODULE}".format(tpe=power_spec_arg, spec=power_spec_file, TOP_MODULE=top_module))

        # Set options pre-elaboration
        verbose_append("set_db leakage_power_effort medium")
        verbose_append("set_db lp_insert_clock_gating true")

        if self.level == FlowLevel.RTL:
            # Elaborate the design
            verbose_append("elaborate {TOP_MODULE}".format(TOP_MODULE=top_module))
        elif self.level == FlowLevel.SYN:
            # Read in the synthesized netlist
            verbose_append("read_netlist {}".format(" ".join(self.input_files)))

            # Read in the post-synth SDCs
            verbose_append("read_sdc {}".format(self.sdc))

        return True

    def read_stimulus(self) -> bool:
        verbose_append = self.verbose_append

        top_module = self.get_setting("power.inputs.top_module")
        tb_name = self.tb_name
        # Replace . to / formatting in case argument passed from sim tool
        tb_dut = self.tb_dut.replace(".", "/")

        # Generate average power report for all waveforms
        waveforms = self.waveforms
        for i, waveform in enumerate(waveforms):
            verbose_append("read_stimulus -file {WAVE} -dut_instance {TB}/{DUT} -alias {WAVE_NAME}_{NUM} -append".format(WAVE=waveform, TB=tb_name, DUT=tb_dut, WAVE_NAME=os.path.basename(waveform), NUM=i))

        # Generate Specified and Custom Reports
        reports = self.get_power_report_configs()

        for i, report in enumerate(reports):
            waveform = os.path.basename(report.waveform_path)

            read_stim_cmd = "read_stimulus -file {WAVE_PATH} -dut_instance {TB}/{DUT} -append".format(WAVE_PATH=report.waveform_path, TB=tb_name, DUT=tb_dut)

            if report.start_time:
                read_stim_cmd += " -start {STIME}".format(STIME=report.start_time.value_in_units("ns"))

            if report.end_time:
                read_stim_cmd += " -end {ETIME}".format(ETIME=report.end_time.value_in_units("ns"))

            if report.toggle_signal:
                if report.num_toggles:
                    read_stim_cmd += " -cycles {NUM} {SIGNAL}".format(NUM=report.num_toggles, SIGNAL=report.toggle_signal)
                else:
                    self.logger.error("Must specify the number of toggles if the toggle signal is specified.")
                    return False

            if report.frame_count:
                read_stim_cmd += " -frame_count {FRAME_COUNT}".format(FRAME_COUNT=report.frame_count)


            read_stim_cmd += " -alias report_{WAVE}_{NUM}".format(WAVE=waveform, NUM=i)

            verbose_append(read_stim_cmd)

        saifs = self.get_setting("power.inputs.saifs")
        for saif in saifs:
            saif_basename = os.path.basename(saif)
            verbose_append("read_stimulus {SAIF} -dut_instance {TB}/{DUT} -format saif -alias {NAME} -append".format(SAIF=saif, TB=tb_name, DUT=tb_dut, NAME=saif_basename))

        return True


    def synthesize_design(self) -> bool:
        verbose_append = self.verbose_append

        if self.level == FlowLevel.RTL:
            # Generate and read the SDCs
            sdc_files = self.generate_sdc_files()  # type: List[str]
            verbose_append("read_sdc {}".format(" ".join(sdc_files)))
            verbose_append("syn_power -effort medium")

        return True


    def compute_power(self) -> bool:
        verbose_append = self.verbose_append

        verbose_append("compute_power -mode time_based")

        return True


    def report_power(self) -> bool:
        verbose_append = self.verbose_append

        top_module = self.get_setting("power.inputs.top_module")
        tb_name = self.tb_name
        # Replace . to / formatting in case argument passed from sim tool
        tb_dut = self.tb_dut.replace(".", "/")

        for i, wave in enumerate(self.waveforms):
            verbose_append("report_power -stims {WAVE}_{NUM} -indent_inst -unit mW -append -out waveforms.report".format(WAVE=os.path.basename(wave), NUM=i))

        reports = self.get_power_report_configs()

        for i, report in enumerate(reports):
            waveform = os.path.basename(report.waveform_path)

            module_str = ""
            if report.module:
                module_str = " -module {MODULE}".format(MODULE=report.module)

            levels_str = ""
            if report.levels:
                levels_str = " -levels {LEVELS}".format(LEVELS=report.levels)

            stim_alias = "report_{WAVE}_{NUM}".format(WAVE=waveform, NUM=i)
            if report.report_name:
                report_name = report.report_name
            else:
                report_name = stim_alias + ".report"

            verbose_append("report_power -frames [get_sdb_frames {STIM_ALIAS}] -collate none -cols total -by_hierarchy{MODULE}{LEVELS} -indent_inst -unit mW -out {REPORT_NAME}".format(
                STIM_ALIAS=stim_alias,
                MODULE=module_str,
                LEVELS=levels_str,
                REPORT_NAME=report_name))

        saifs = self.get_setting("power.inputs.saifs")
        for saif in saifs:
            saif_basename = os.path.basename(saif)
            verbose_append("report_power -stims {SAIF} -indent_inst -unit mW -out {SAIF}.report".format(SAIF=saif_basename))

        return True

    def run_joules(self) -> bool:
        verbose_append = self.verbose_append

        """Close out the power script and run Joules"""
        # Quit Joules
        verbose_append("exit")

        # Create power analysis script
        joules_tcl_filename = os.path.join(self.run_dir, "joules.tcl")
        self.write_contents_to_path("\n".join(self.output), joules_tcl_filename)

        # Build args
        args = [
            self.get_setting("power.joules.joules_bin"),
            "-files", joules_tcl_filename,
            "-common_ui"
        ]

        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False

        self.run_executable(args, cwd=self.run_dir)

        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

        return True



tool = Joules



#  hammer-vlsi plugin for Synopsys IC Validator.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerToolStep
from hammer.vlsi import HammerLVSTool
from hammer.common.synopsys import SynopsysTool
from hammer.logging import HammerVLSILogging
from hammer.utils import HammerFiletype, get_filetype
import hammer.tech as hammer_tech

from typing import Dict, List, Optional

import os
import textwrap


class ICVLVS(HammerLVSTool, SynopsysTool):

    def tool_config_prefix(self) -> str:
        return "lvs.icv"

    def erc_results_pre_waived(self) -> Dict[str, int]:
        return {}

    def lvs_results(self) -> List[str]:
        return []

    def fill_outputs(self) -> bool:
        return True

    # TODO: placeholder empty step
    def empty_step(self) -> bool:
        return True

    @property
    def steps(self) -> List[HammerToolStep]:
        steps = [self.generate_lvs_run_file, self.generate_lvs_args_file]  # TODO: LVS steps require multiple runs of the tool how do we support this?
        return self.make_steps_from_methods(steps)

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_icv()

    def globally_waived_erc_rules(self) -> List[str]:
        return []

    def run_icv(self) -> bool:
        self.create_enter_script()

        # translate all spice & verilog netlists with vue_nettran
        self.generate_top_icv_file()
        # generate the hcells file if needed
        self.generate_hcells_file()

        # set the command arguments
        args = [
            self.get_setting("lvs.icv.icv_lvs_bin"),
            "-64"]  # always want to be in 64-bit mode
        if self.version() >= self.version_number("R-2020.09"):
            args.extend(["-host_init", str(self.get_setting("vlsi.core.max_threads"))])
        else:
            args.append("-dp{}".format(self.get_setting("vlsi.core.max_threads")))
        args.extend([
            "-clf",
            self.lvs_args_file,
            "-vue",  # needed to view results in VUE
            "-verbose" # get more than % complete
        ])
        args.append(self.lvs_run_file)

        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False
        self.run_executable(args, cwd=self.run_dir)  # TODO: check for errors and deal with them
        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

        # TODO: check that lvs run was successful

        # Create view_lvs script & icvwb macro script file
        # See the README for how this works
        os.makedirs(self.generated_scripts_dir, exist_ok=True)

        with open(self.icvwb_macrofile, "w") as f:
            # Open socket
            f.write("user_socket open 0\n")
            # Layer mapping
            layerprops_file = self.get_setting("synopsys.layerprops")
            if layerprops_file is not None:
                f.writelines(l for l in open(layerprops_file))

        with open(self.view_lvs_script, "w") as f:
            f.write("""
        cd {run_dir}
        source ./enter
        # Start Synopsys IC Validator WorkBench and wait for port to open before starting VUE
        {icvwb} -socket {port} -run {macrofile} {gds} &
        while ! nc -z localhost {port}; do
            sleep 0.1
        done
        {icv_vue} -64 -load {results} -lay icwb -layArgs Port {port}
            """.format(
                run_dir=self.run_dir,
                icvwb=self.get_setting("lvs.icv.icvwb_bin"),
                port=self.get_setting("lvs.icv.icvwb_port"),
                macrofile=self.icvwb_macrofile,
                gds=self.layout_file,
                icv_vue=self.get_setting("lvs.icv.icv_vue_bin"),
                results=self.lvs_results_db
            ))
        os.chmod(self.view_lvs_script, 0o755)

        return True

    @property
    def hcells_file(self) -> str:
        return os.path.join(self.run_dir, "hcells")

    def generate_hcells_file(self) -> None:
        with open(self.hcells_file, "w") as f:
            f.write("")
            # TODO

    def generate_top_icv_file(self) -> None:
        library_spice_files = self.technology.read_libs([hammer_tech.filters.spice_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
        ilms = list(map(lambda x: x.netlist, self.ilms))  # type: List[str]

        all_files = library_spice_files + self.schematic_files + ilms
        spice_files = list(filter(lambda x: get_filetype(x) is HammerFiletype.SPICE, all_files))
        verilog_files = list(filter(lambda x: get_filetype(x) is HammerFiletype.VERILOG, all_files))
        unmatched = set(all_files).symmetric_difference(set(spice_files + verilog_files))
        if unmatched:
            raise NotImplementedError("Unsupported netlist type for files: " + str(unmatched))

        args = [self.get_setting("lvs.icv.icv_nettran_bin"), "-sp-autoDetectBusdelimiter", "FIRST", "-sp"]
        args.extend(spice_files)
        args.extend(["-verilog"] + verilog_files)
        args.extend(["-outName", self.converted_icv_file])
        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False
        self.run_executable(args, cwd=self.run_dir)  # TODO: check for errors and deal with them
        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

    def generate_lvs_run_file(self) -> bool:
        """ Generate the LVS run file self.lvs_run_file and fill its contents """
        with open(self.lvs_run_file, "w") as f:
            f.write(self.header.replace("#","//") + "\n\n")
            f.write(self.get_additional_lvs_text())
            # Include paths to all supplied decks
            for rule in self.get_lvs_decks():
                f.write("#include <{}>\n".format(rule.path))
        return True

    def generate_lvs_args_file(self) -> bool:
        """ Generate the LVS args file self.lvs_args_file and fill its contents """
        with open(self.lvs_args_file, "w") as f:
            f.write(textwrap.dedent("""
            # Generated by HAMMER
            -i {gds}\
            -c {top}\
            -f GDSII\
            -s {sch}\
            -sf ICV
            """).format(
                gds=self.layout_file,
                top=self.top_module,
                sch=self.converted_icv_file
            )
            )
            # Symbolic variables to set via command. Can also use #define <var> <value> in additional_lvs_text.
            defines = self.get_setting("lvs.icv.defines")  # type: List[Dict[str, str]]
            assert isinstance(defines, list)
            if len(defines) > 0:
                # Most comprehensive way of covering all List[Dict] possibilities
                f.write(" -D " + " -D ".join(map(lambda x: " -D ".join("=".join(_) for _ in x.items()), defines)))

            # Preprocessor directories to include
            include_dirs = self.get_setting("lvs.icv.include_dirs")  # type: List[str]
            assert isinstance(include_dirs, list)
            if len(include_dirs) > 0:
                f.write(" -I " + " -I ".join(include_dirs))

            # Config runset file
            config_rs = self.get_setting("lvs.icv.config_runset")  # type: Optional[str]
            if config_rs is not None:
                f.write(" -runset_config " + config_rs)
        return True

    @property
    def generated_scripts_dir(self) -> str:
        return os.path.join(self.run_dir, "generated-scripts")

    @property
    def view_lvs_script(self) -> str:
        return os.path.join(self.generated_scripts_dir, "view_lvs")

    @property
    def icvwb_macrofile(self) -> str:
        return os.path.join(self.generated_scripts_dir, "icvwb_macrofile")

    @property
    def lvs_run_file(self) -> str:
        return os.path.join(self.run_dir, "lvs_run_file")

    @property
    def lvs_args_file(self) -> str:
        return os.path.join(self.run_dir, "lvs_args_file")

    @property
    def erc_results_file(self) -> str:
        return os.path.join(self.run_dir, self.top_module + ".LAYOUT_ERRORS")

    @property
    def lvs_results_db(self) -> str:
        return os.path.join(self.run_dir, self.top_module + ".vue")

    @property
    def lvs_results_file(self) -> str:
        return os.path.join(self.run_dir, self.top_module + ".LVS_ERRORS")

    @property
    def converted_icv_file(self) -> str:
        return os.path.join(self.run_dir, "{top}.icv".format(top=self.top_module))

    @property
    def env_vars(self) -> Dict[str, str]:
        """
        Get the list of environment variables required for this tool.
        Note to subclasses: remember to include variables from super().env_vars!
        """
        result = dict(super().env_vars)
        result.update({
            "ICV_HOME_DIR": self.get_setting("lvs.icv.ICV_HOME_DIR"),
            "PATH": "{path}:{icv_home}/bin/LINUX.64".format(path=os.environ.copy()["PATH"], icv_home=self.get_setting("lvs.icv.ICV_HOME_DIR"))
        })
        return result


tool = ICVLVS



#  Hammer logging code.
#
#  See LICENSE for licence details.

from functools import reduce
from enum import Enum
from typing import Callable, Iterable, List, NamedTuple, Type, Optional

class Level(Enum):
    """
    Logging levels.
    """
    # Explanation of logging levels:
    # DEBUG - for debugging only, too verbose for general use
    # INFO - general informational messages (e.g. "starting synthesis")
    # WARNING - things the user should check out (e.g. a setting uses something very usual)
    # ERROR - something gone wrong but the process can still continue (e.g. synthesis run failed), though the user should definitely check it out.
    # FATAL - an error which will abort the process immediately without warning (e.g. assertion failure)
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3
    FATAL = 4


# Message including additional metadata such as level and context.
FullMessage = NamedTuple('FullMessage', [
    ('message', str),
    ('level', Level),
    ('context', List[str])
])


# Need a way to bind the callbacks to the class...
def with_default_callbacks(cls):
    # TODO: think about how to remove default callbacks
    cls.add_callback(cls.callback_print)
    cls.add_callback(cls.callback_buffering)
    return cls


class HammerVLSIFileLogger:
    """A file logger for HammerVLSILogging."""

    def __init__(self, output_path: str, format_msg_callback: Optional[Callable[[FullMessage], str]] = None) -> None:
        """
        Create a new file logger.

        :param output_path: Output path of the logger.
        :param format_msg_callback: Optional callback to run to build the message. None to use HammerVLSILogging.build_log_message.
        """
        self._file = open(output_path, "a")
        self._format_msg_callback = format_msg_callback

    def __enter__(self):
        return self

    def close(self) -> None:
        """
        Close this file logger.
        """
        self._file.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def callback(self) -> Callable[[FullMessage], None]:
        """Get the callback for HammerVLSILogging.add_callback."""

        def file_callback(fullmessage: FullMessage) -> None:
            if self._format_msg_callback is not None:
                self._file.write(self._format_msg_callback(fullmessage) + "\n")
            else:
                self._file.write(HammerVLSILogging.build_log_message(fullmessage) + "\n")

        return file_callback


@with_default_callbacks
class HammerVLSILogging:
    """Singleton which handles logging in hammer-vlsi.

    This class is generally not intended to be used directly for logging, but through HammerVLSILoggingContext instead.
    """

    # Enable colour in logging?
    enable_colour = True  # type: bool

    # If buffering is enabled, instead of printing right away, we will store it
    # into a buffer for later retrieval.
    enable_buffering = False  # type: bool

    # Enable printing the tag (e.g. "[synthesis] ...).
    enable_tag = True  # type: bool

    # Various escape characters for colour output.
    COLOUR_BLUE = "\033[96m"
    COLOUR_GREY = "\033[37m"
    COLOUR_YELLOW = "\033[33m"
    COLOUR_RED = "\033[91m"
    COLOUR_RED_BG = "\033[101m"
    # Restore the default terminal colour.
    COLOUR_CLEAR = "\033[0m"

    # Some default callback implementations.
    @classmethod
    def callback_print(cls, fullmessage: FullMessage) -> None:
        """Default callback which prints a colour message."""
        print(cls.build_message(fullmessage))

    output_buffer = []  # type: List[str]

    @classmethod
    def callback_buffering(cls, fullmessage: FullMessage) -> None:
        """Get the current contents of the logging buffer and clear it."""
        if not cls.enable_buffering:
            return
        cls.output_buffer.append(cls.build_message(fullmessage))

    @classmethod
    def build_log_message(cls, fullmessage: FullMessage) -> str:
        """Build a plain message for logs, without colour."""
        template = "{context} {level}: {message}"

        return template.format(context=cls.get_tag(fullmessage.context), level=fullmessage.level,
                               message=fullmessage.message)

    # List of callbacks to call for logging.
    callbacks = []  # type: List[Callable[[FullMessage], None]]

    @classmethod
    def clear_callbacks(cls) -> None:
        """Clear the list of callbacks."""
        cls.callbacks = []

    @classmethod
    def add_callback(cls, callback: Callable[[FullMessage], None]) -> None:
        """Add a callback."""
        cls.callbacks.append(callback)

    @classmethod
    def context(cls, new_context: str = "") -> "HammerVLSILoggingContext":
        """
        Create a new context.

        :param new_context: Context name. Leave blank to get the global context.
        """
        if new_context == "":
            return HammerVLSILoggingContext([], cls)
        else:
            return HammerVLSILoggingContext([new_context], cls)

    @classmethod
    def get_colour_escape(cls, level: Level) -> str:
        """Colour table to translate level -> colour in logging."""
        table = {
            Level.DEBUG: cls.COLOUR_GREY,
            Level.INFO: cls.COLOUR_BLUE,
            Level.WARNING: cls.COLOUR_YELLOW,
            Level.ERROR: cls.COLOUR_RED,
            Level.FATAL: cls.COLOUR_RED_BG
        }
        if level in table:
            return table[level]
        else:
            return ""

    @classmethod
    def log(cls, fullmessage: FullMessage) -> None:
        """
        Log the given message at the given level in the given context.
        """
        for callback in cls.callbacks:
            callback(fullmessage)

    @classmethod
    def build_message(cls, fullmessage: FullMessage) -> str:
        """Build a colour message."""
        message = fullmessage.message
        level = fullmessage.level
        context = fullmessage.context

        context_tag = cls.get_tag(context)  # type: str

        output = ""  # type: str
        output += cls.get_colour_escape(level) if cls.enable_colour else ""
        if cls.enable_tag and context_tag != "":
            output += context_tag + " "
        output += message
        output += cls.COLOUR_CLEAR if cls.enable_colour else ""

        return output

    @staticmethod
    def get_tag(context: List[str]) -> str:
        """Helper function to get the tag for outputting a message given a context."""
        if len(context) > 0:
            return str(reduce(lambda a, b: a + " " + b, map(lambda x: "[%s]" % (x), context)))
        else:
            return "[<global>]"

    @classmethod
    def get_buffer(cls) -> Iterable[str]:
        """Get the current contents of the logging buffer and clear it."""
        if not cls.enable_buffering:
            raise ValueError("Buffering is not enabled")
        output = list(cls.output_buffer)
        cls.output_buffer = []
        return output


class HammerVLSILoggingContext:
    """
    Logging interface to hammer-vlsi which contains a context (list of strings denoting hierarchy where the log occurred).
    e.g. ["synthesis", "subprocess run-synthesis"]
    """

    def __init__(self, context: List[str], logging_class: Type[HammerVLSILogging]) -> None:
        """
        Create a new interface with the given context.
        """
        self._context = context  # type: List[str]
        self.logging_class = logging_class  # type: Type[HammerVLSILogging]

    def context(self, new_context: str) -> "HammerVLSILoggingContext":
        """
        Create a new subcontext from this context.
        """
        context2 = list(self._context)
        context2.append(new_context)
        return HammerVLSILoggingContext(context2, self.logging_class)

    def debug(self, message: str) -> None:
        """Create an debug-level log message."""
        return self.log(message, Level.DEBUG)

    def info(self, message: str) -> None:
        """Create an info-level log message."""
        return self.log(message, Level.INFO)

    def warning(self, message: str) -> None:
        """Create an warning-level log message."""
        return self.log(message, Level.WARNING)

    def error(self, message: str) -> None:
        """Create an error-level log message."""
        return self.log(message, Level.ERROR)

    def fatal(self, message: str) -> None:
        """Create an fatal-level log message."""
        return self.log(message, Level.FATAL)

    def log(self, message: str, level: Level) -> None:
        return self.logging_class.log(FullMessage(message, level, self._context))



#  hooks.py
#  Classes and functions related to Hammer hooks.
#
#  See LICENSE for licence details.

from enum import Enum
from typing import Callable, NamedTuple, Optional, TYPE_CHECKING

__all__ = ['HammerStepFunction', 'HammerToolStep', 'HookLocation', 'HammerToolHookAction']

# Necessary for mypy to learn about HammerTool.
if TYPE_CHECKING:
    from .hammer_tool import HammerTool  # pylint: disable=unused-import

HammerStepFunction = Callable[['HammerTool'], bool]

HammerToolStep = NamedTuple('HammerToolStep', [
    # Function to call to execute this step
    ('func', HammerStepFunction),
    # Name of the step
    ('name', str)
])

# Specify step to start/stop
HammerStartStopStep = NamedTuple('HammerStartStopStep', [
    # Name of the step
    ('step', Optional[str]),
    # Whether it is inclusive
    ('inclusive', bool)
])

# Where to insert/replace the given step.
# Persistent steps always execute depending on from/after/to/until steps.
class HookLocation(Enum):
    InsertPreStep = 1
    InsertPostStep = 2
    ReplaceStep = 10
    ResumePreStep = 20
    ResumePostStep = 21
    PausePreStep = 30
    PausePostStep = 31
    PersistentStep = 40
    PersistentPreStep = 41
    PersistentPostStep = 42

# An hook action. Actions can insert new steps before or after an existing one,
# or replace an existing stage.
# Note: hook actions are executed in the order provided.
HammerToolHookAction = NamedTuple('HammerToolHookAction', [
    # Where/what to do
    ('location', HookLocation),
    # Target step to insert before/after or replace
    ('target_name', str),
    # Step to insert/replace
    ('step', Optional[HammerToolStep])
])



#  hammer-vlsi plugin for generic PCB deliverables. This can generate
#  footprints, schematic symbols, and CSVs for visualizing the bump map.
#
#  See LICENSE for licence details.

import os
import datetime
import re
from typing import List, Dict, Tuple
from decimal import Decimal
from hammer.utils import um2mm

from hammer.vlsi import HammerPCBDeliverableTool, HammerToolStep, BumpsPinNamingScheme, BumpsDefinition, BumpAssignment

class GenericPCBDeliverableTool(HammerPCBDeliverableTool):

    def tool_config_prefix(self) -> str:
        return "pcb.generic"

    def version_number(self, version: str) -> int:
        return 1

    def fill_outputs(self) -> bool:
        self.output_footprints = [self.output_footprint_filename]
        self.output_schematic_symbols = [self.output_schematic_symbol_filename]
        return True

    @property
    def steps(self) -> List[HammerToolStep]:
        steps = [
            self.create_footprint_csv,
            self.create_footprint,
            self.create_bom_builder_pindata_txt,
            self.create_schematic_symbols
        ]
        return self.make_steps_from_methods(steps)

    @property
    def output_footprint_filename(self) -> str:
        # change this when supporting other types
        return os.path.join(self.run_dir, "{top}-pads.d".format(top=self.top_module))

    @property
    def output_footprint_csv_filename(self) -> str:
        return os.path.join(self.run_dir, "{top}-pads.csv".format(top=self.top_module))

    @property
    def output_schematic_symbol_filename(self) -> str:
        # change this when supporting other types
        return os.path.join(self.run_dir, "{top}-symbol.csv".format(top=self.top_module))

    @property
    def output_bom_builder_pindata_filename(self) -> str:
        return os.path.join(self.run_dir, "{top}-pindata.txt".format(top=self.top_module))

    @property
    def footprint_name(self) -> str:
        return "TESTCHIP-{}".format(self.top_module)

    @property
    def bump_pad_opening_diameter(self) -> Decimal:
        """
        Get the diameter of the PCB footprint pad opening in post-shrink um.
        """
        diameter = self.get_setting("technology.pcb.bump_pad_opening_diameter")
        if diameter is None:
            raise ValueError("Must provide a technology.pcb.bump_pad_opening_diameter if generating PCB deliverables.")
        else:
            return Decimal(str(diameter))

    @property
    def bump_pad_metal_diameter(self) -> Decimal:
        """
        Get the diameter of the PCB footprint pad metal in post-shrink um.
        """
        diameter = self.get_setting("technology.pcb.bump_pad_metal_diameter")
        if diameter is None:
            raise ValueError("Must provide a technology.pcb.bump_pad_metal_diameter if generating PCB deliverables.")
        else:
            return Decimal(str(diameter))

    def get_bumps_and_assignments(self) -> Tuple[BumpsDefinition, List[BumpAssignment]]:
        """
        Helper method to get non-Optional versions of the bumps definition and bumps assignment list.
        """
        bumps = self.get_bumps()
        if bumps is None:
            raise ValueError("There must be a bumps definition in order to generate to generate PCB deliverables!")
        assert isinstance(bumps, BumpsDefinition)

        assignments = bumps.assignments
        if assignments is None:
            raise ValueError("There must be at least 1 bump in order to generate PCB deliverables!")
        assert isinstance(assignments, list)

        if bumps.x < 1 or bumps.y < 1 or len(bumps.assignments) < 1:
            raise ValueError("There must be at least 1 bump in order to generate PCB deliverables!")

        return (bumps, assignments)

    def create_bom_builder_pindata_txt(self) -> bool:
        """
        Creates a PINDATA.txt file that is consumed by the BOMBuilder tool to create a package.
        BOMBuilder is part of the PartSync tool suite: http://www.partsync.com/
        """

        bumps, assignments = self.get_bumps_and_assignments()
        naming_scheme = self.naming_scheme

        # Sort the bumps
        sorted_assignments = naming_scheme.sort_by_name(bumps, assignments)

        # Use post-shrink pitch
        pitch_x = self.technology.get_post_shrink_length(bumps.pitch_x)
        pitch_y = self.technology.get_post_shrink_length(bumps.pitch_y)

        # There is no meaningful verion for this file
        # Set the units to metric (mm)
        output = "VER 0.0\nUNIT M\n"
        x_offset = ((1 - bumps.x) * pitch_x) / 2  # type: Decimal
        y_offset = ((1 - bumps.y) * pitch_y) / 2  # type: Decimal
        for bump in sorted_assignments:
            # note that the flip-chip mirroring happens here
            name = naming_scheme.name_bump(bumps, bump)
            x = um2mm(Decimal(str(bumps.x - bump.x)) * pitch_x + x_offset, 3)
            y = um2mm(Decimal(str(bump.y - 1)) * pitch_y + y_offset, 3)
            # Fields in order (with valid options):
            # Name
            # X position (mm)
            # Y position (mm)
            # Shape (RND or RECT)
            # Width (mm)
            # Height (mm)
            # Side (TOP or BOT)
            # Connection offset X (mm)
            # Connection offset Y (mm)
            # Lead type (0 = signal pin, 1 = mounting hole, 2 = shield)
            # Hole lock (T = true, F = false)
            # Plated/Not Plated (P = Plated, N = Not plated)
            output += "{name:<6} {x:>10.3f} {y:>10.3f} RND {w:0.3f} {w:0.3f} NONE SMT TOP 0 0 0 F P\n".format(name=name, x=x, y=y, w=um2mm(self.bump_pad_metal_diameter, 3))

        with open(self.output_bom_builder_pindata_filename, "w") as f:
            f.write(output)

        return True

    def create_footprint_csv(self) -> bool:
        """
        Create a CSV of the bump map to be easily read into your spreadsheet editor of choice.
        This will fail silently with non-integer bump assignment coordinates (they'll be ignored).

        :return: True if successful
        """

        bumps, assignments = self.get_bumps_and_assignments()

        output = ""
        for y in range(bumps.y,0,-1):
            for x in range(1,bumps.x+1):
                # This is a pretty inefficient algorithm for this task
                matches = [z for z in bumps.assignments if z.x == x and z.y == y]
                if len(matches) > 0:
                    output += str(matches[0].name)
                output += ","
            output += "\n"

        with open(self.output_footprint_csv_filename, "w") as f:
            f.write(output)

        return True

    def create_footprint(self) -> bool:
        """
        Create the footprint type requested by pcb.generic.footprint_type. Currently only supports PADS-V9,
        and does not correctly output soldermask or solderpaste openings.

        :return: True if successful
        """
        footprint_type = str(self.get_setting("pcb.generic.footprint_type"))
        if footprint_type != "PADS-V9":
            raise NotImplementedError("Unsupported footprint type: {}".format(footprint_type))

        bumps, assignments = self.get_bumps_and_assignments()

        # Use post-shrink pitch
        pitch_x = self.technology.get_post_shrink_length(bumps.pitch_x)
        pitch_y = self.technology.get_post_shrink_length(bumps.pitch_y)

        # Here is a good ref for the PADS V9 format:
        # ftp://ftp.freecalypso.org/pub/CAD/PADS/pdfdocs/Plib_ASCII.pdf

        # for now, let's make the outline 4 pitches larger than the bump array
        outline_width = ((bumps.x + 3) * pitch_x)
        outline_height = ((bumps.y + 3) * pitch_y)

        output = "*PADS-LIBRARY-PCB-DECALS-V9*\n\n"
        # Fields in order from left to right
        # name
        # coordinate units type (M=metric, mm)
        # origin x
        # origin y
        # number of attributes (0)
        # number of labels (0)
        # number of drawing pieces (1: the outline)
        # number of free text strings (0)
        # number of terminals (number of bumps)
        # number of pad stack definitions (1)
        # maxlayers mode (0 = standard)
        output += "{top} M 0 0 0 0 1 0 {terminals} 1 0\n".format(top=self.footprint_name, terminals=len(bumps.assignments))

        # create the timestamp
        output += datetime.datetime.now().strftime("TIMESTAMP %Y.%m.%d.%H.%M.%S\n")

        # create the outline
        # 6 points, 0.1 mm wide, layer 1, solid line
        output += "OPEN 6 0.1 1 -1\n"

        # we'll dog-ear the outline to indicate the reference bump in the top left
        # x and y are in mm, not um
        output += "{x} {y}\n".format(x=um2mm(-outline_width/2, 3),           y=um2mm(-outline_height/2, 3))
        output += "{x} {y}\n".format(x=um2mm(-outline_width/2, 3),           y=um2mm( outline_height/2 - 2*pitch_y, 3))
        output += "{x} {y}\n".format(x=um2mm(-outline_width/2 + 2*pitch_x, 3), y=um2mm( outline_height/2, 3))
        output += "{x} {y}\n".format(x=um2mm( outline_width/2, 3),           y=um2mm( outline_height/2, 3))
        output += "{x} {y}\n".format(x=um2mm( outline_width/2, 3),           y=um2mm(-outline_height/2, 3))
        output += "{x} {y}\n".format(x=um2mm(-outline_width/2, 3),           y=um2mm(-outline_height/2, 3))

        # create all of the terminals
        x_offset = ((1 - bumps.x) * pitch_x) / 2  # type: Decimal
        y_offset = ((1 - bumps.y) * pitch_y) / 2  # type: Decimal
        naming_scheme = self.naming_scheme
        for bump in bumps.assignments:
            # note that the flip-chip mirroring happens here
            x = um2mm(Decimal(str(bumps.x - bump.x)) * pitch_x + x_offset, 3)
            y = um2mm(Decimal(str(bump.y - 1)) * pitch_y + y_offset, 3)
            label = naming_scheme.name_bump(bumps, bump)
            output += "T{x} {y} {x} {y} {label}\n".format(x=x, y=y, label=label)

        # create the pad definition
        # 0=default pad
        # 3=num layer lines
        # P=plated
        # 0=drill diameter (0=no drill)
        output += "PAD 0 3 P 0\n"
        # -2=top layer
        # diameter
        # R=round pad
        output += "-2 {dia} R\n".format(dia=um2mm(self.bump_pad_opening_diameter, 3))
        # -1=all inner layers
        # 0=no pad
        # R=round pad
        output += "-1 0 R\n"
        # 0=bottom layer
        # 0=no pad
        # R=round pad
        output += "0 0 R\n"

        output += "\n*END*"

        with open(self.output_footprint_filename, "w") as f:
            f.write(output)

        return True

    def create_schematic_symbols(self) -> bool:
        """
        Create the schematic symbol type requested by pcb.generic.schematic_symbol_type. Currently only supports
        AltiumCSV, which is a format Altium allows to be copy-and-pasted into the schematic symbol wizard.
        This will divide the pins into groups based on their "group" field.

        :return: True if successful
        """
        schematic_symbol_type = str(self.get_setting("pcb.generic.schematic_symbol_type"))
        if schematic_symbol_type != "AltiumCSV":
            raise NotImplementedError("Unsupported schematic symbool type: {}".format(schematic_symbol_type))

        bumps, assignments = self.get_bumps_and_assignments()

        naming_scheme = self.naming_scheme
        groups = {}  # type: Dict[str, List[BumpAssignment]]

        for a in assignments:
            group = a.group
            if a.group is None:
                # TODO make me smarter
                if a.name in ["VDD", "VSS"]:
                    group = a.name
                else:
                    group = "ungrouped"
            assert isinstance(group, str)


            if group not in groups:
                groups[group] = [a]
            else:
                # Cap each group at 50
                while group in groups and len(groups[group]) >= 50:
                    match = re.match(r"^(.*)_(\d+)$", group)
                    if match:
                        old = int(match.group(2))
                        group = match.group(1) + "_{new}".format(new=old+1)
                    else:
                        group = group + "_1"

                if group not in groups:
                    groups[group] = [a]
                else:
                    groups[group].append(a)

        group_names = sorted([n for n in groups])
        num_rows = max([len(v) for k,v in groups.items()])

        output = ""
        # Build the headers
        for group in group_names:
            output += "{group},,,,,,,,".format(group=group)
        output += "\n"
        # Now the column headers
        for group in group_names:
            output += "Position,Group,Display Name,Designator,Electrical Type,Description,Side,,"
        output += "\n"

        for i in range(num_rows):
            for group in group_names:
                if len(groups[group]) > i:
                    bump = groups[group][i]
                    # TODO(johnwright) Use the Verilog to make Passive actually have a meaningful direction
                    output += "{pos},,{name},{des},Passive,,Left,,".format(pos=i+1,name=bump.name,des=naming_scheme.name_bump(bumps, bump))
                else:
                    output += ",,,,,,,,"
            output += "\n"

        with open(self.output_schematic_symbol_filename, "w") as f:
            f.write(output)


        return True


tool = GenericPCBDeliverableTool



#  mocksram_generator.py
#  Mock SRAM Generator tool for testing.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerSRAMGeneratorTool, DummyHammerTool, \
        SRAMParameters, MMMCCorner
from hammer.tech import ExtraLibrary, Library
from typing import List, Dict

class MockSRAMGenerator(HammerSRAMGeneratorTool, DummyHammerTool):
    def generate_sram(self, params: SRAMParameters, corner: MMMCCorner) -> ExtraLibrary:
        return ExtraLibrary(prefix=None,
               library=Library(gds_file="sram{d}x{w}_{v}V_{c}C.gds".format(
                   d=params.depth, w=params.width,
                   v=corner.voltage.value, c=corner.temp.value)))

tool = MockSRAMGenerator
